;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: laccept.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Ken Forbus, Johan de Kleer, John Everett
;;;;   Created: May 1, 1986
;;;;  Modified: everett on Sat Jul 6 20:25:48 1996
;;;;   Purpose: LTRE Acceptance tests
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------


(in-package :LTRE)

(rule-file "ACCEPT")

(defun test-ltre (&optional (dbg nil))
  (in-ltre (create-ltre "Debugging LTRE" :dbg-flags dbg))
  (format t "~%Testing database/LTMS link...")
  (test-datums)
  (format t "~%Testing LTMS...")
  (test-clauses)
  (format t "~%Testing Rule system...")
  (test-rules))

(defun test-datums ()
  (assert! 'foo 'testing)
  (unless (true? 'foo) (error "Fact installation glitch"))
  (assert! '(:NOT bar) 'testing)
  (unless (false? 'bar) (error "Negation glitch"))
  (format t "~%~%****Passed datums test~%"))

(defun test-clauses ()
   (assert! '(:OR a b) 'case-split)
   (assert! '(:IMPLIES a c) 'why-not?)
   (assume! '(:IMPLIES c d) 'what-the-heck)
   (assume! '(:NOT b) 'for-fun)
   (unless (true? 'd) (error "Propagation glitch"))
   (retract! '(:NOT b) 'for-fun)
   (unless (unknown? 'd) (error "Retraction glitch"))
   (assume! '(:NOT b) 'for-fun)
   (unless (true? 'd) (error "Unouting glitch"))
   (retract! '(:IMPLIES c d) 'what-the-heck)
   (unless (unknown? 'd) (error "Retraction glitch 2"))
   (assume!'(:IMPLIES c d) 'what-the-heck)
   (unless (true? 'd) (error "Unouting glitch 2"))
   (format t "~%~%****Passed clauses test~%"))

(defun test-rules ()
   (eval `(rule ((:TRUE (foo ?xt) :VAR ?f1)
                 (:TRUE (bar ?yt) :VAR ?f2))
             (rassert! (:IMPLIES (:AND ?f1 ?f2) (mumble ?xt ?yt)) 'hack)))
   (eval `(rule ((:INTERN (foo ?xi) :VAR ?f1)
                 (:INTERN (bar ?yi) :VAR ?f2))
             (rassert! (:IMPLIES (:AND ?f1 ?f2) (grumble ?xi ?yi)) 'hack)))
   (format t "~&Referencing (FOO 1)~%")
   (referent '(foo 1) t)
   (referent '(bar 1) t)
   (run-rules)
   (unless (referent '(grumble 1 1) nil) (error "Intern triggering failure"))
   (when (referent '(mumble 1 1) nil) (error "Premature triggering"))
   (format t "Assuming (FOO 1) and (NOT (BAR 1))~%")
   (assume! '(foo 1) 'why-not?)
   (assume! '(:not (bar 1)) 'monkeywrench)
   (run-rules)
   (when (true? '(mumble 1 1)) (error "Badly conditioned triggering - 1"))
   (format t "Retracting (NOT (BAR 1)) with wrong informant~%")
   (retract! '(:not (bar 1)) 'tweak)
   (unless (false? '(bar 1)) (error "Retraction with wrong informant"))
   (format t "Retracting (NOT (BAR 1))~%")
   (retract! '(:not (bar 1)) 'monkeywrench)
   (run-rules)
   (when (true? '(mumble 1 1)) (error "Badly conditioned triggering - 2"))
   (format t "Assuming (BAR 1)~%")
   (assume! '(bar 1) 'why)
   (run-rules)
   (unless (true? '(mumble 1 1)) (error "Badly conditioned triggering - 3"))
   (format t "Assuming (FOO 2)~%")
   (assume! '(foo 2) 'go-for-it)
   (run-rules)
   (unless (true? '(mumble 2 1)) (error "Rule chaining failure"))
   (format t "Assuming (BAR 2)~%")
   (assume! '(bar 2) 'alternate)
   (run-rules)
   (unless (true? '(mumble 1 2)) (error "Subrule spawning failure - 1"))
   (unless (true? '(mumble 2 2)) (error "Subrule spawning failure - 2"))
   (format t "~%~%****Passed rules test~%"))

(defun show-r1 ()
   (pprint 
      (macroexpand
         '(rule ((:TRUE (foo ?xt) :VAR ?f1)
                 (:TRUE (bar ?yt) :VAR ?f2))
             (rassert! (:IMPLIES (:AND ?f1 ?f2) (mumble ?xt ?yt)) 'hack)))))

#| TEST RULE EXPANSIONS

(rule ((:INTERN (foo ?xi) :VAR ?f1)
       (:TRUE (fooble ?zzi) :VAR ?f10)
       (:intern (bar ?yi) :VAR ?f2))
   (rassert! (:IMPLIES (:AND ?f1 ?f2 ?f10) (grumble ?xi ?yi)) 'hack))

;;;RULE 1 (true conditions on triggers)
(progn
   (defun |-(foo ?x)-144| (trigger-node ?f1 ?x)
      (cond ((true-node? trigger-node) 
            (insert-rule (get-dbclass 'bar) 
               #'(lambda (p)
                   (|-(bar ?y)-142| p ?f1 ?x))
               #'(lambda (trigger-node ?y ?f2)
                   (|-(bar ?y)-143| trigger-node ?y ?f2 ?f1 ?x))))
         (t (setf (rule-bindings *current-rule*) (list trigger-node ?f1 ?x)) 
            (push *current-rule* (tms-node-true-rules trigger-node)))))
   
   (defun |-(foo ?x)-141| (p)
      (if (and (consp p) 
               (equal 'foo (car p))
               (consp (cdr p)) 
               (null (cdr (cdr p))))
         (values t (list p (car (cdr p))))))
   
   (defun |-(bar ?y)-143| (trigger-node ?f2 ?y ?f1 ?x)
      (cond ((true-node? trigger-node) 
             (rassert! (:IMPLIES (:AND ?f1 ?f2) (mumble ?x ?y)) 'hack))
            (t (setf (rule-bindings *current-rule*) 
                     (list trigger-node ?f2 ?y ?f1 ?x))
               (push *current-rule* 
                  (tms-node-true-rules trigger-node)))))
   
   (defun |-(bar ?y)-142| (p ?f1 ?x)
      (if (and (consp p) 
               (equal 'bar (car p))
               (consp (cdr p))
               (null (cdr (cdr p))))
         (values t (list p (car (cdr p))))))
   
   (insert-rule (get-dbclass 'foo) 
      #'|-(foo ?x)-141| 
      #'(lambda (trigger-node ?x ?f1)
           (|-(foo ?x)-144| trigger-node ?x ?f1))))

(progn
   (defun |-(foo ?x)-148| (trigger-node ?f1 ?x)
      (push *current-rule* (tms-node-intern-rules trigger-node))
      (format t "~&current rule is: ~a~%" *current-rule*)
      (insert-rule (get-dbclass 'bar) 
         #'(lambda (p)
              (|-(bar ?y)-146| p ?f1 ?x))
         #'(lambda (trigger-node ?y ?f2)
              (|-(bar ?y)-147| trigger-node ?y ?f2 ?f1 ?x))))
   
   (defun |-(foo ?x)-145| (p)
      (if (and (consp p) 
               (equal 'foo (car p))
               (consp (cdr p))
               (null (cdr (cdr p))))
         (values t (list p (car (cdr p))))))
   
   (defun |-(bar ?y)-147| (trigger-node ?f2 ?y ?f1 ?x)
      (push *current-rule* (tms-node-intern-rules trigger-node))
      (format t "~&current rule is: ~a~%" *current-rule*)
      (rassert! (:IMPLIES (:AND ?f1 ?f2) (grumble ?x ?y)) 'hack))
   
   (defun |-(bar ?y)-146| (p ?f1 ?x)
      (if (and (consp p) 
               (equal 'bar (car p)) 
               (consp (cdr p))
               (null (cdr (cdr p))))
         (values t (list p (car (cdr p))))))
   
   (insert-rule (get-dbclass 'foo) #'|-(foo ?x)-145| #'|-(foo ?x)-148|))

(rule ((:TRUE (foo ?xt) :VAR ?f1)
       (:TRUE (bar ?yt) :VAR ?f2)
       (:TRUE (baz ?zt) :VAR ?f3))
   (rassert! (:IMPLIES (:AND ?f1 ?f2 ?f3)
              (mumble ?xt ?yt ?zt)) 'hack))


|#